# Emmy version 1.0.0
# EmmyDDManager.py
# by Antoine Rauzy
# Copyrights (c) 2019 AltaRica Association
#
# This file is part of Emmy, a pedagogical stochastic simulator.
# Emmy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Emmy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Emmy.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creators
# 3) Literals
# 4) Nodes
# 5) Logical operations
# 6) Minimal cutsets
# 7) Counts and probabilities

# 1) Imported modules
# -------------------

from S2MLCore import *

from EmmyDDNode import *
from EmmyDDUniqueTable import *
from EmmyDDCache import *

# 2) Creators
# -----------

class EmmyDDManager:
  NOT = 1
  OR  = 2
  AND = 3
  EQ  = 4
  DF  = 5
  IF  = 6
  WITHIN = 7
  MCS = 8
  WITHOUT = 9
  Pr = 10

  def __init__(self, uniqueTableSize, cacheSize):
    from EmmyFDSLiteral import EmmyFDSLiteral
    self._fdsLiterals = []
    self._universe = S2MLUniverse()
    self._leaves = dict()
    self._uniqueTable = EmmyDDUniqueTable(uniqueTableSize)
    self._cache = EmmyDDCache(cacheSize)
    self._nextNodeNumber = 0
    self._leafNone = self.NewLeaf(self._universe.NewNone())
    self._leafFalse = self.NewLeaf(self._universe.NewBoolean(False))
    # self._leafFalse = self._leafNone
    self._leafTrue = self.NewLeaf(self._universe.NewBoolean(True))

  def Reset(self):
    self._uniqueTable.Reset()
    self._cache.Reset()
    self._fdsLiterals.clear()

# 3) Literals
# -----------

  def GetFDSLiterals(self):
    return self._fdsLiterals

  def GetFDSLiteral(self, index):
    if index<0 or index>=len(self._fdsLiterals):
      return None
    return self._fdsLiterals[index]

  def SetFDSLiteral(self, index, fdsLiteral):
    while len(self._fdsLiterals)<index:
      self._fdsLiterals.append(None)
    self._fdsLiterals.append(fdsLiteral)
    fdsLiteral.SetDDIndex(index)

# 4) Nodes
# --------

  def IsLeafNone(self, node):
    return self._leafNone==node

  def IsLeafFalse(self, node):
    return self._leafFalse==node

  def IsLeafTrue(self, node):
    return self._leafTrue==node

  def NewLeafNone(self):
    return self._leafNone

  def NewLeafFalse(self):
    return self._leafFalse

  def NewLeafTrue(self):
    return self._leafTrue

  def NewLeaf(self, constant):
    actualConstant = self._universe.NewConstant(constant)
    leaf = self._leaves.get(actualConstant, None)
    if leaf==None:
      leaf = EmmyDDNode.NewLeaf(self._nextNodeNumber, actualConstant)
      self._leaves[actualConstant] = leaf
      self._nextNodeNumber += 1
    return leaf

  def IsLeaf(self, node, constant):
    actualConstant = node.GetConstant()
    if actualConstant==None:
      return False
    return actualConstant.IsEq(constant)

  def NewNode(self, index, thenChild, elseChild):
    from EmmyFDSLiteral import EmmyFDSLiteral
    if self.IsLeafNone(thenChild):
      return elseChild
    fdsLiteral = self.GetFDSLiteral(index)
    if fdsLiteral.GetPosition()==EmmyFDSLiteral.FIRST:
      node = elseChild
      expectedIndex = index+1
      allEqual = True
      while True:
        if node.IsLeaf() or node.GetIndex()!=expectedIndex or node.GetThenChild()!=thenChild:
          allEqual = False
          break
        fdsLiteral = self.GetFDSLiteral(expectedIndex)
        if fdsLiteral.GetPosition()==EmmyFDSLiteral.LAST:
          break
        node = node.GetElseChild()
        expectedIndex += 1
      if allEqual:
        return thenChild
    node = self._uniqueTable.LookForNode(index, thenChild, elseChild)
    if node==None:  
      node = self._uniqueTable.NewNode(self._nextNodeNumber, index, thenChild, elseChild)
      self._nextNodeNumber += 1
    return node

# 5) Logical operations
# ---------------------

  def ComputeNot(self, node):
    if self.IsLeafNone(node):
      return self.NewLeafNone()
    if self.IsLeafFalse(node):
      return self.NewLeafTrue()
    if self.IsLeafTrue(node):
      return self.NewLeafFalse()
    result = self._cache.LookUpUnaryOperation(EmmyDDManager.NOT, node)
    if result!=None:
      return result
    result1 = self.ComputeNot(node.GetThenChild())
    result0 = self.ComputeNot(node.GetElseChild())
    result = self.NewNode(node.GetIndex(), result1, result0)
    self._cache.InsertUnaryOperation(EmmyDDManager.NOT, node, result)
    return result

  def ComputeWithin(self, node, constants):
    actualConstants = []
    for constant in constants:
      actualConstant = self._universe.NewConstant(constant)
      actualConstants.append(actualConstant)
    rootNode = self._ComputeWithin(node, actualConstants)
    self._cache.RemoveOperationsOfCode(EmmyDDManager.WITHIN)
    return rootNode

  def _ComputeWithin(self, node, constants):
    if self.IsLeafNone(node):
      return self.NewLeafNone()
    if node.IsLeaf():
      for constant in constants:
        if node.GetConstant()==constant:
          return self.NewLeafTrue()
      return self.NewLeafFalse()
    result = self._cache.LookUpUnaryOperation(EmmyDDManager.WITHIN, node)
    if result!=None:
      return result
    result1 = self._ComputeWithin(node.GetThenChild(), constants)
    result0 = self._ComputeWithin(node.GetElseChild(), constants)
    result = self.NewNode(node.GetIndex(), result1, result0)
    self._cache.InsertUnaryOperation(EmmyDDManager.WITHIN, node, result)
    return result

  def ComputeOr(self, node1, node2):
    return self.ComputeBinaryCommutativeOperation(EmmyDDManager.OR, node1, node2)

  def ComputeAnd(self, node1, node2):
    return self.ComputeBinaryCommutativeOperation(EmmyDDManager.AND, node1, node2)

  def ComputeEq(self, node1, node2):
    return self.ComputeBinaryCommutativeOperation(EmmyDDManager.EQ, node1, node2)

  def ComputeDf(self, node1, node2):
    return self.ComputeBinaryCommutativeOperation(EmmyDDManager.DF, node1, node2)

  def ComputeBinaryCommutativeOperation(self, operationCode, node1, node2):
    result = self.ComputeTerminalValue(operationCode, node1, node2)
    if result!=None:
      return result
    if node1.IsLeaf() or (node2.IsNode() and node1.GetIndex()>node2.GetIndex()):
      node3 = node2
      node2 = node1
      node1 = node3
    result = self._cache.LookUpBinaryOperation(operationCode, node1, node2)
    if result!=None:
      return result
    if node2.IsLeaf() or (node2.IsNode() and node1.GetIndex()<node2.GetIndex()):
      result0 = self.ComputeBinaryCommutativeOperation(operationCode, node1.GetElseChild(), node2)
      node3 = self.GetOtherwiseChild(node1, node2)
      result1 = self.ComputeBinaryCommutativeOperation(operationCode, node1.GetThenChild(), node3)
    else:
      result0 = self.ComputeBinaryCommutativeOperation(operationCode, node1.GetElseChild(), node2.GetElseChild())
      result1 = self.ComputeBinaryCommutativeOperation(operationCode, node1.GetThenChild(), node2.GetThenChild())
    result = self.NewNode(node1.GetIndex(), result1, result0)
    self._cache.InsertBinaryOperation(operationCode, node1, node2, result)
    return result

  def ComputeTerminalValue(self, operationCode, node1, node2):
    if self.IsLeafNone(node1) or self.IsLeafNone(node2):
      return self.NewLeafNone()
    if operationCode==EmmyDDManager.OR:
      if self.IsLeafTrue(node1) or self.IsLeafTrue(node2):
        return self.NewLeafTrue()
      if self.IsLeafFalse(node1) and self.IsLeafFalse(node2):
        return self.NewLeafFalse()
      return None
    if operationCode==EmmyDDManager.AND:
      if self.IsLeafFalse(node1) or self.IsLeafFalse(node2):
        return self.NewLeafFalse()
      if self.IsLeafTrue(node1) and self.IsLeafTrue(node2):
        return self.NewLeafTrue()
      return None
    if operationCode==EmmyDDManager.EQ:
      if node1==node2:
        return self.NewLeafTrue()
      if node1.IsLeaf() and node2.IsLeaf():
        return self.NewLeafFalse()
      return None
    if operationCode==EmmyDDManager.DF:
      if node1==node2:
        return self.NewLeafFalse()
      if node1.IsLeaf() and node2.IsLeaf():
        return self.NewLeafTrue()
      return None
    return None

  def GetOtherwiseChild(self, node1, node2):
    fdsLiteral1 = self._fdsLiterals[node1.GetIndex()]
    fdsVariable1 = fdsLiteral1.GetFDSVariable()
    result = node2
    while True:
      if result.IsLeaf():
        break
      fdsLiteral2 = self._fdsLiterals[result.GetIndex()]
      fdsVariable2 = fdsLiteral2.GetFDSVariable()
      if fdsVariable2!=fdsVariable1:
        break
      result = result.GetElseChild()
    return result

  def ComputeIfThenElse(self, node1, node2, node3):
    if self.IsLeafNone(node1):
      return self.NewLeafNone()
    if self.IsLeafTrue(node1):
      return node2
    if self.IsLeafFalse(node1):
      return node3
    result = self._cache.LookUpTernaryOperation(EmmyDDManager.IF, node1, node2, node3)
    if result!=None:
      return result
    if node2.IsNode():
      if node2.GetIndex()<node1.GetIndex():
        if node3.IsNode():
          if node3.GetIndex()<node2.GetIndex():
            # node2.GetIndex()<node1.GetIndex() and node3.GetIndex()<node2.GetIndex()
            node4 = self.GetOtherwiseChild(node3, node1)
            node5 = self.GetOtherwiseChild(node3, node2)
            result1 = self.ComputeIfThenElse(node4, node5, node3.GetThenChild())
            result0 = self.ComputeIfThenElse(node1, node2, node3.GetElseChild())
            result = self.NewNode(node3.GetIndex(), result1, result0)
          elif node3.GetIndex()>node2.GetIndex():
            # node2.GetIndex()<node1.GetIndex() and node2.GetIndex()<node3.GetIndex()
            node4 = self.GetOtherwiseChild(node2, node1)
            node5 = self.GetOtherwiseChild(node2, node3)
            result1 = self.ComputeIfThenElse(node4, node2.GetThenChild(), node5)
            result0 = self.ComputeIfThenElse(node1, node2.GetElseChild(), node3)
            result = self.NewNode(node2.GetIndex(), result1, result0)
          else:
            # node2.GetIndex()<node1.GetIndex() and node2.GetIndex()==node3.GetIndex()
            node4 = self.GetOtherwiseChild(node2, node1)
            result1 = self.ComputeIfThenElse(node4, node2.GetThenChild(), node3.GetThenChild())
            result0 = self.ComputeIfThenElse(node1, node2.GetElseChild(), node3.GetElseChild())
            result = self.NewNode(node2.GetIndex(), result1, result0)
        else:
          # node2.GetIndex()<node1.GetIndex() and node3.IsLeaf()
          node4 = self.GetOtherwiseChild(node2, node1)
          result1 = self.ComputeIfThenElse(node4, node2.GetThenChild(), node3)
          result0 = self.ComputeIfThenElse(node1, node2.GetElseChild(), node3)
          result = self.NewNode(node2.GetIndex(), result1, result0)
      elif node2.GetIndex()>node1.GetIndex():
        if node3.IsNode():
          if node3.GetIndex()<node1.GetIndex():
            # node1.GetIndex()<node2.GetIndex() and node3.GetIndex()<node1.GetIndex()
            node4 = self.GetOtherwiseChild(node3, node1)
            node5 = self.GetOtherwiseChild(node3, node2)
            result1 = self.ComputeIfThenElse(node4, node5, node3.GetThenChild())
            result0 = self.ComputeIfThenElse(node1, node2, node3.GetElseChild())
            result = self.NewNode(node3.GetIndex(), result1, result0)
          elif node3.GetIndex()>node1.GetIndex():
            # node1.GetIndex()<node2.GetIndex() and node1.GetIndex()<node3.GetIndex()
            node4 = self.GetOtherwiseChild(node1, node2)
            node5 = self.GetOtherwiseChild(node1, node3)
            result1 = self.ComputeIfThenElse(node1.GetThenChild(), node4, node5)
            result0 = self.ComputeIfThenElse(node1.GetElseChild(), node2, node3)
            result = self.NewNode(node1.GetIndex(), result1, result0)
          else: # node3.GetIndex()==node1.GetIndex()
            # node1.GetIndex()<node2.GetIndex() and node1.GetIndex()==node3.GetIndex()
            node4 = self.GetOtherwiseChild(node1, node2)
            result1 = self.ComputeIfThenElse(node1.GetThenChild(), node4, node3.GetThenChild())
            result0 = self.ComputeIfThenElse(node1.GetElseChild(), node2, node3.GetElseChild())
            result = self.NewNode(node1.GetIndex(), result1, result0)
        else: # node3.IsLeaf()
          # node1.GetIndex()<node2.GetIndex() and node3.IsLeaf()
          node4 = self.GetOtherwiseChild(node1, node2)
          result1 = self.ComputeIfThenElse(node1.GetThenChild(), node4, node3)
          result0 = self.ComputeIfThenElse(node1.GetElseChild(), node2, node3)
          result = self.NewNode(node1.GetIndex(), result1, result0)
      else:
        if node3.IsNode():
          if node3.GetIndex()<node1.GetIndex():
            # node1.GetIndex()==node2.GetIndex() and node3.GetIndex()<node1.GetIndex()
            node4 = self.GetOtherwiseChild(node3, node1)
            node5 = self.GetOtherwiseChild(node3, node2)
            result1 = self.ComputeIfThenElse(node4, node5, node3.GetThenChild())
            result0 = self.ComputeIfThenElse(node1, node2, node3.GetElseChild())
            result = self.NewNode(node3.GetIndex(), result1, result0)
          elif node3.GetIndex()>node1.GetIndex():
            # node1.GetIndex()==node2.GetIndex() and node1.GetIndex()<node3.GetIndex()
            node4 = self.GetOtherwiseChild(node1, node3)
            result1 = self.ComputeIfThenElse(node1.GetThenChild(), node2.GetThenChild(), node4)
            result0 = self.ComputeIfThenElse(node1.GetElseChild(), node2.GetElseChild(), node3)
            result = self.NewNode(node1.GetIndex(), result1, result0)
          else:
            # node1.GetIndex()==node2.GetIndex() and node1.GetIndex()==node3.GetIndex()
            result1 = self.ComputeIfThenElse(node1.GetThenChild(), node2.GetThenChild(), node3.GetThenChild())
            result0 = self.ComputeIfThenElse(node1.GetElseChild(), node2.GetElseChild(), node3.GetElseChild())
            result = self.NewNode(node1.GetIndex(), result1, result0)
        else:
          # node1.GetIndex()==node2.GetIndex() and node3.IsLeaf()
          result1 = self.ComputeIfThenElse(node1.GetThenChild(), node2.GetThenChild(), node3)
          result0 = self.ComputeIfThenElse(node1.GetElseChild(), node2.GetElseChild(), node3)
          result = self.NewNode(node1.GetIndex(), result1, result0)
    else:
      if node3.IsNode():
        if node3.GetIndex()<node1.GetIndex():
          # node2.IsLeaf() and node3.GetIndex()<node1.GetIndex()
          node4 = self.GetOtherwiseChild(node3, node1)
          result1 = self.ComputeIfThenElse(node4, node2, node3.GetThenChild())
          result0 = self.ComputeIfThenElse(node1, node2, node3.GetElseChild())
          result = self.NewNode(node3.GetIndex(), result1, result0)
        elif node3.GetIndex()>node1.GetIndex():
          # node2.IsLeaf() and node1.GetIndex()<node3.GetIndex()
          node4 = self.GetOtherwiseChild(node1, node3)
          result1 = self.ComputeIfThenElse(node1.GetThenChild(), node2, node4)
          result0 = self.ComputeIfThenElse(node1.GetElseChild(), node2, node3)
          result = self.NewNode(node1.GetIndex(), result1, result0)
        else:
          # node2.IsLeaf() and node1.GetIndex()==node3.GetIndex()
          result1 = self.ComputeIfThenElse(node1.GetThenChild(), node2, node3.GetThenChild())
          result0 = self.ComputeIfThenElse(node1.GetElseChild(), node2, node3.GetElseChild())
          result = self.NewNode(node1.GetIndex(), result1, result0)
      else:
        # node2.IsLeaf() and node3.IsLeaf()
        result1 = self.ComputeIfThenElse(node1.GetThenChild(), node2, node3)
        result0 = self.ComputeIfThenElse(node1.GetElseChild(), node2, node3)
        result = self.NewNode(node1.GetIndex(), result1, result0)
    self._cache.InsertTernaryOperation(EmmyDDManager.IF, node1, node2, node3, result)
    return result

# 6) Minimal cutsets
# ------------------

  def ExtractMinimalCutsets(self, node):
    if self.IsLeafTrue(node):
      return self.NewLeafTrue()
    if node.IsLeaf():
      return self.NewLeafNone()
    fdsLiteral = self.GetFDSLiteral(node.GetIndex())
    fdsConstant = fdsLiteral.GetFDSConstant()
    if fdsConstant.IsFDSLeastElement():
      return self.ExtractMinimalCutsets(node.GetThenChild())
    result = self._cache.LookUpUnaryOperation(EmmyDDManager.MCS, node)
    if result!=None:
      return result
    result0 = self.ExtractMinimalCutsets(node.GetElseChild())
    result2 = self.ExtractMinimalCutsets(node.GetThenChild())
    result1 = self.Prune(fdsLiteral, result2, result0)
    result = self.NewNode(node.GetIndex(), result1, result0)
    self._cache.InsertUnaryOperation(EmmyDDManager.MCS, node, result)
    return result

  def Prune(self, fdsLiteral1, node1, node2):
    if self.IsLeafNone(node2):
      return node1
    if self.IsLeafTrue(node2):
      return self.NewLeafNone()
    fdsLiteral2 = self.GetFDSLiteral(node2.GetIndex())
    if fdsLiteral1.GetFDSVariable()==fdsLiteral2.GetFDSVariable():
      fdsConstant1 = fdsLiteral1.GetFDSConstant()
      fdsConstant2 = fdsLiteral2.GetFDSConstant()
      if fdsConstant1.IsFDSBigger(fdsConstant2):
        result = self.Without(node1, node2.GetThenChild())
        result = self.Prune(fdsLiteral1, result, node2.GetElseChild())
      else:
        result = self.Prune(fdsLiteral1, node1, node2.GetElseChild())
    else:
      result = self.Without(node1, node2)
    return result

  def Without(self, node1, node2):
    if self.IsLeafNone(node1) or self.IsLeafTrue(node2):
      return self.NewLeafNone()
    if self.IsLeafTrue(node1) or self.IsLeafNone(node2):
      return node1
    if node1.GetIndex()>node2.GetIndex():
      return self.Without(node1, node2.GetElseChild())
    result = self._cache.LookUpBinaryOperation(EmmyDDManager.WITHOUT, node1, node2)
    if result!=None:
      return result
    fdsLiteral = self.GetFDSLiteral(node1.GetIndex())
    if node1.GetIndex()<node2.GetIndex():
      result0 = self.Without(node1.GetElseChild(), node2)
      result1 = self.Prune(fdsLiteral, node1.GetThenChild(), node2)
    else:
      result0 = self.Without(node1.GetElseChild(), node2.GetElseChild())
      result1 = self.Without(node1.GetThenChild(), node2.GetThenChild())
      result1 = self.Prune(fdsLiteral, result1, node2.GetElseChild())
      # result1 = self.Prune(fdsLiteral, result1, result0)
    result = self.NewNode(node1.GetIndex(), result1, result0)
    self._cache.InsertBinaryOperation(EmmyDDManager.WITHOUT, node1, node2, result)
    return result

# 7) Counts and probabilities
# ---------------------------

# 7.1) Number of nodes
# --------------------

# 7.2) Size
# ---------

  def ComputeSize(self, node):
    size = self._ComputeSize(node)
    node.ResetFlags()
    return size

  def _ComputeSize(self, node):
    if node.GetFlag()!=0:
      return (0, 0)
    node.SetFlag(1)
    if node.IsLeaf():
      return (1, 0)
    (numberOfLeaves1, numberOfNodes1) = self._ComputeSize(node.GetThenChild())
    (numberOfLeaves0, numberOfNodes0) = self._ComputeSize(node.GetElseChild())
    return (numberOfLeaves1+numberOfLeaves0, numberOfNodes1+numberOfNodes0+1)

# 7.3) Orders of products
# -----------------------

  def CollectOrdersOfProducts(self, node):
    orders = []
    self._CollectOrdersOfProducts(node, 0, orders)
    return orders

  def _CollectOrdersOfProducts(self, node, order, orders):
    if self.IsLeafTrue(node):
      while len(orders)<=order:
        orders.append(0)
      orders[order] += 1
      return
    if node.IsLeaf():
      return
    self._CollectOrdersOfProducts(node.GetThenChild(), order+1, orders)
    self._CollectOrdersOfProducts(node.GetElseChild(), order, orders)

  def ComputeProbability(self, node):
    Pr = self._ComputeProbability(node)
    self._cache.RemoveOperationsOfCode(EmmyDDManager.Pr)
    return Pr

  def _ComputeProbability(self, node):
    if self.IsLeafTrue(node):
      return 1.0
    if node.IsLeaf():
      return 0.0
    Pr = self._cache.LookUpUnaryOperation(EmmyDDManager.Pr, node)
    if Pr!=None:
      return Pr
    fdsLiteral = self._fdsLiterals[node.GetIndex()]
    Pr1 = self._ComputeProbability(node.GetThenChild())
    Pr0 = self._ComputeProbability(node.GetElseChild())
    Pr = fdsLiteral.GetProbability()*Pr1 + Pr0
    self._cache.InsertUnaryOperation(EmmyDDManager.Pr, node, Pr)
    return Pr

# 8) Rat hole
# -----------

  def DumpFDSLiteral(self, fdsLiteral):
    variable = fdsLiteral.GetVariable()
    constant = fdsLiteral.GetConstant()
    return "%s=%s@%d" % (variable.GetIdentifier(), constant.ToString(), fdsLiteral.GetDDIndex())

  def DumpNode(self, node):
    from EmmyVariable import EmmyVariable
    text = "<%d: " % node.GetNumber()
    if node.IsLeaf():
      constant = node.GetConstant()
      text += "%s" % constant.ToString()
    else:
      fdsLiteral = self._fdsLiterals[node.GetIndex()]
      text += self.DumpFDSLiteral(fdsLiteral)
      text += ", %d, %d" % (node._thenChild.GetNumber(), node._elseChild.GetNumber())
    text += ">"
    return text