# Emmy version 1.0.0
# EmmyChecker.py
# by Antoine Rauzy
# Copyrights (c) 2019 AltaRica Association
#
# This file is part of Emmy, a pedagogical stochastic simulator.
# Emmy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Emmy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Emmy.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Undefined references
# 4) Circular definitions
# 5) Types
# 6) Tool specific checks

# 1) Imported modules
# -------------------

from S2MLCore import *

from EmmyModels import *

# 2) Creator
# ----------

class EmmyChecker(S2MLChecker):
  def __init__(self, engine):
    S2MLChecker.__init__(self, engine)

# 3) Undefined references
# -----------------------

  def CheckReferencesInBlock(self, block):
    for element in block.GetElements():
      if element.GetType()==EmmyType.BLOCK:
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CheckReferencesInBlock(element)
        self._checkedBlock = previousCheckedBlock
      elif element.GetType()==EmmyType.VARIABLE:
        self._checkedElement = element
        self._CheckReferencesInAttributesOfElement(element)
        self._checkedElement = None
      elif element.GetType()==EmmyType.PARAMETER:
        self._checkedElement = element
        self._CheckReferencesInStochasticExpression(element.GetDefinition())
        self._checkedElement = None
      elif element.GetType()==EmmyType.PROBABILITY_DECLARATION:
        self._checkedElement = element
        self._CheckReferencesInProbabilityDeclaration(element)
        self._checkedElement = None
      elif element.GetType()==EmmyType.ASSIGNMENT:
        self._checkedElement = element
        self._CheckReferencesInAssignment(element)
        self._checkedElement = None
      elif element.GetType()==EmmyType.OBSERVER:
        self._checkedElement = element
        self._CheckReferencesInObserver(element)
        self._checkedElement = None
    return self._errors

  def _CheckReferencesInProbabilityDeclaration(self, declaration):
    variablePath = declaration.GetVariablePath()
    variable = self._checkedBlock.AccessNamedElement(variablePath)
    if variable==None:
      self._RaiseError("In block %s, undeclared variable %s" % (self._checkedBlock.GetIdentifier(), variablePath.ToString()))
      return
    declaration.SetVariable(variable)
    distribution = declaration.GetDistribution()
    self._CheckReferencesInStochasticExpression(distribution)

  def _CheckReferencesInAssignment(self, assignment):
    variablePath = assignment.GetVariablePath()
    variable = self._checkedBlock.AccessNamedElement(variablePath)
    if variable==None:
      self._RaiseError("In block %s, undeclared variable %s" % (self._checkedBlock.GetIdentifier(), variablePath.ToString()))
      return
    assignment.SetVariable(variable)
    expression = assignment.GetExpression()
    self._CheckReferencesInStochasticExpression(expression)

  def _CheckReferencesInObserver(self, observer):
    definition = observer.GetDefinition()
    self._CheckReferencesInStochasticExpression(definition)

# 4) Circular definitions
# -----------------------
# Circular definition are checked by the rewriter

# 5) Types
# --------
# Not implemented yet

# 6) Tool specific checks
# -----------------------

  def ToolSpecificCheckModel(self, model):
    self.CheckCategoriesOfVariables(model)

  def CheckCategoriesOfVariables(self, block):
    for element in block.GetElements():
      if element.GetType()==EmmyType.BLOCK:
        previousCheckedBlock = self._checkedBlock
        self._checkedBlock = element
        self.CheckCategoriesOfVariables(element)
        self._checkedBlock = previousCheckedBlock
      elif element.GetType()==EmmyType.VARIABLE:
        self.CheckCategoryOfVariable(element)
      elif element.GetType()==EmmyType.ASSIGNMENT:
        self.CheckCategoryOfLeftMemberOfAssignment(element)
    return self._errors

  def CheckCategoryOfVariable(self, variable):
    expression = variable.GetAttribute("init")
    if expression!=None:
      variable.SetCategory(EmmyVariable.STATE)
      return
    expression = variable.GetAttribute("reset")
    if expression!=None:
      variable.SetCategory(EmmyVariable.FLOW)
      return
    self.RaiseError("Attribute init/reset is missing for variable %s" % variable.GetIdentifier())

  def CheckCategoryOfLeftMemberOfAssignment(self, assignment):
    variablePath = assignment.GetVariablePath()
    variable = self._checkedBlock.AccessNamedElement(variablePath)
    if variable.GetCategory()==EmmyVariable.STATE:
      self.RaiseError("Variable %s declared as state but occurs as left member of an equation" % variable.GetIdentifier())
      return
    if variable.GetDefinition()!=None:
      self.RaiseError("Variable %s is defined twice" % variable.GetIdentifier())
      return
    variable.SetDefinition(assignment.GetExpression())
