# Emmy version 1.0.0
# EmmyFlattener.py
# by Antoine Rauzy
# Copyrights (c) 2019 AltaRica Association
#
# This file is part of Emmy, a pedagogical stochastic simulator.
# Emmy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Emmy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Emmy.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Flattener
#   3.1) Creator
#   3.2) Instantiation of elements
#   3.3) Flattening of elements

# 1) Imported modules
# -------------------

from S2MLCore import *

from EmmyModels import *
from EmmyPrinter import *

# 3) Flattener
# ------------

# 3.1) Creator
# ------------

class EmmyFlattener(S2MLFlattener):
  def __init__(self, engine):
    S2MLFlattener.__init__(self, engine)

# 3.2) Instantiation of elements
# -----------------------------

  def NewModel(self):
    return EmmyModel()

  def InstantiateElement(self, action):
    sourceElement = action.GetDatum(0)
    if sourceElement.GetType()==EmmyType.DOMAIN_DECLARATION:
      status = self._InstantiateDomainDeclaration(action)
    elif sourceElement.GetType()==EmmyType.VARIABLE:
      status = self._InstantiateVariable(action)
    elif sourceElement.GetType()==EmmyType.PROBABILITY_DECLARATION:
      status = self._InstantiateProbabilityDeclaration(action)
    elif sourceElement.GetType()==EmmyType.ASSIGNMENT:
      status = self._InstantiateAssignment(action)
    elif sourceElement.GetType()==EmmyType.OBSERVER:
      status = self._InstantiateObserver(action)
    else:
      status = S2MLFlattener.InstantiateElement(self, action)
    return status

  def _InstantiateDomainDeclaration(self, action):
    sourceDeclaration = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceIdentifier = sourceDeclaration.GetIdentifier()
    targetIdentifier = sourceIdentifier
    sourceDomain = sourceDeclaration.GetDomain()
    targetDomain = self._CloneDomain(sourceDomain)
    targetDeclaration = self._targetModel.NewDomainDeclaration(targetIdentifier, targetDomain, targetContainer)
    for sourceConstantOrder in sourceDeclaration.GetConstantOrders():
      sourceConstant1 = sourceConstantOrder.GetConstant1()
      sourceConstant2 = sourceConstantOrder.GetConstant2()
      targetConstant1 = self._targetModel.NewConstant(sourceConstant1)
      targetConstant2 = self._targetModel.NewConstant(sourceConstant2)
      targetDeclaration.AppendConstantOrder(targetConstant1, targetConstant2)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateVariable(self, action):
    sourceVariable = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetPath = sourceVariable.GetPath()
    targetElement = targetContainer.AccessNamedElement(targetPath)
    if targetElement==None:
      if targetPath.IsIdentifier():
        sourceDomain = sourceVariable.GetDomain()
        targetDomain = self._CloneDomain(sourceDomain)
        targetVariable = self._targetModel.NewVariable(targetDomain, targetPath, targetContainer)
      else:
        absoluteSourcePath = sourceVariable.GetAbsolutePath()
        self.RaiseError("Variable %s declared outside of its parent block\n" % absoluteSourcePath.ToString())
        return S2MLFlattenerAction.FAILED
    elif targetElement.GetType()==EmmyType.VARIABLE:
      targetVariable = targetElement # self._targetModel.NewVariable(targetDomainIdentifier, targetPath, targetContainer)
      sourceDomain = sourceVariable.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      targetVariable.SetDomain(targetDomain)
    else:
      absoluteSourcePath = sourceVariable.GetAbsolutePath()
      self.RaiseError("Type mismatch for variable %s\n" % absoluteSourcePath.ToString())
      return S2MLFlattenerAction.FAILED
    contextualizedPath = targetContainer.GetContextualizedAbsolutePath()
    for name in sourceVariable.GetAttributeNames():
      sourceExpression = sourceVariable.GetAttribute(name)
      targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, contextualizedPath)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetVariable.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateProbabilityDeclaration(self, action):
    sourceDeclaration = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceVariablePath = sourceDeclaration.GetVariablePath()
    targetVariable = targetContainer.AccessNamedElement(sourceVariablePath)
    if targetVariable==None or targetVariable.GetType()!=EmmyType.VARIABLE:
      container = sourceDeclaration.GetContainer()
      absoluteSourcePath = container.GetAbsolutePath()
      self.RaiseError("Undeclared variable %s in probability distribution declared in %s\n" % (sourceVariablePath.ToString(), absoluteSourcePath.ToString()))
      return S2MLFlattenerAction.FAILED
    contextualizedPath = targetContainer.GetContextualizedAbsolutePath()
    if sourceVariablePath.IsLocal():
      targetVariablePath = S2MLPath.Copy(sourceVariablePath)
    else:
      targetVariablePath = S2MLPath.Merge(contextualizedPath, sourceVariablePath)
    sourceConstant = sourceDeclaration.GetConstant()
    targetConstant = self._targetModel.NewConstant(sourceConstant)
    sourceDistribution = sourceDeclaration.GetDistribution()
    targetDistribution = self._ContextualizedCopyOfStochasticExpression(sourceDistribution, contextualizedPath)
    if targetDistribution==None:
      return S2MLFlattenerAction.FAILED
    targetDeclaration = self._targetModel.NewProbabilityDeclaration(targetVariablePath, targetConstant, targetDistribution, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateAssignment(self, action):
    sourceAssignment = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    sourceVariablePath = sourceAssignment.GetVariablePath()
    targetElement = targetContainer.AccessNamedElement(sourceVariablePath)
    if targetElement==None:
      self.RaiseError("Undeclared variable %s in block %s\n" % (sourceVariablePath.ToString(), targetContainer.GetAbsolutePath().ToString()))
      return S2MLFlattenerAction.FAILED
    if targetElement.GetType()!=EmmyType.VARIABLE:
      self.RaiseError("Type mismatch for variable %s in block %s\n" % (sourceVariablePath.ToString(), targetContainer.GetAbsolutePath().ToString()))
      return S2MLFlattenerAction.FAILED
    contextualizedPath = targetContainer.GetContextualizedAbsolutePath()
    if sourceVariablePath.IsLocal():
      targetVariablePath = S2MLPath.Copy(sourceVariablePath)
    else:
      targetVariablePath = S2MLPath.Merge(contextualizedPath, sourceVariablePath)
    sourceExpression = sourceAssignment.GetExpression()
    targetExpression = self._ContextualizedCopyOfStochasticExpression(sourceExpression, contextualizedPath)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetAssignment = self._targetModel.NewAssignment(targetVariablePath, targetExpression, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _InstantiateObserver(self, action):
    sourceObserver = action.GetDatum(0)
    targetContainer = action.GetDatum(1)
    targetObserverPath = sourceObserver.GetPath()
    targetObserver = targetContainer.AccessNamedElement(targetObserverPath)
    if targetObserver==None:
      if targetObserverPath.IsIdentifier():
        contextualizedPath = targetContainer.GetContextualizedAbsolutePath()
        sourceDefinition = sourceObserver.GetDefinition()
        targetDefinition = self._ContextualizedCopyOfStochasticExpression(sourceDefinition, contextualizedPath)
        targetObserver = self._targetModel.NewObserver(targetObserverPath, targetDefinition, targetContainer)
        targetObserver.Rename(targetObserverPath.ToString())
      else:
        absoluteSourcePath = sourceObserver.GetAbsolutePath()
        self.RaiseError("Observer %s declared outside of its parent block\n" % absoluteSourcePath.ToString())
        return S2MLFlattenerAction.FAILED
    elif targetObserver.GetType()==S2MLType.OBSERVER:
      contextualizedPath = targetContainer.GetContextualizedAbsolutePath()
      sourceDefinition = sourceObserver.GetDefinition()
      targetDefinition = self._ContextualizedCopyOfStochasticExpression(sourceDefinition, contextualizedPath)
      targetObserver.SetDefinition(targetDefinition)
    else:
      absoluteSourcePath = sourceObserver.GetAbsolutePath()
      self.RaiseError("Type mismatch for observer %s\n" % absoluteSourcePath.ToString())
      return S2MLFlattenerAction.FAILED
    return S2MLFlattenerAction.COMPLETED

# 3.7) Flattening of models
# -------------------------

  def FlattenElement(self, action):
    sourceElement = action.GetDatum(1)
    if sourceElement.GetType()==EmmyType.DOMAIN_DECLARATION:
      status = self._FlattenDomainDeclaration(action)
    elif sourceElement.GetType()==EmmyType.VARIABLE:
      status = self._FlattenVariable(action)
    elif sourceElement.GetType()==EmmyType.PROBABILITY_DECLARATION:
      status = self._FlattenProbabilityDeclaration(action)
    elif sourceElement.GetType()==EmmyType.ASSIGNMENT:
      status = self._FlattenAssignment(action)
    elif sourceElement.GetType()==EmmyType.OBSERVER:
      status = self._FlattenObserver(action)
    else:
      status = S2MLFlattener.FlattenElement(self, action)
    return status

  def _FlattenDomainDeclaration(self, action):
    sourceContainer = action.GetDatum(0)
    sourceDeclaration = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    sourceIdentifier = sourceDeclaration.GetIdentifier()
    targetIdentifier = sourceIdentifier
    sourceDomain = sourceDeclaration.GetDomain()
    targetDomain = self._CloneDomain(sourceDomain)
    targetDeclaration = self._targetModel.NewDomainDeclaration(targetIdentifier, targetDomain, targetContainer)
    for sourceConstantOrder in sourceDeclaration.GetConstantOrders():
      sourceConstant1 = sourceConstantOrder.GetConstant1()
      sourceConstant2 = sourceConstantOrder.GetConstant2()
      targetConstant1 = self._targetModel.NewConstant(sourceConstant1)
      targetConstant2 = self._targetModel.NewConstant(sourceConstant2)
      targetDeclaration.AppendConstantOrder(targetConstant1, targetConstant2)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenVariable(self, action):
    sourceContainer = action.GetDatum(0)
    sourceVariable = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    targetPath = sourceVariable.GetAbsolutePath()
    targetPath.RemoveFirstIdentifier()
    targetPath.MakeIdentifier()
    targetVariable = targetContainer.AccessNamedElement(targetPath)
    if targetVariable==None:
      sourceDomain = sourceVariable.GetDomain()
      targetDomain = self._CloneDomain(sourceDomain)
      targetVariable = self._targetModel.NewVariable(targetDomain, targetPath, targetContainer)
      # targetVariable.Rename(targetPath.ToString())
    for name in sourceVariable.GetAttributeNames():
      sourceExpression = sourceVariable.GetAttribute(name)
      targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
      if targetExpression==None:
        return S2MLFlattenerAction.FAILED
      targetVariable.SetAttribute(name, targetExpression)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenProbabilityDeclaration(self, action):
    sourceContainer = action.GetDatum(0)
    sourceDeclaration = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    sourceVariablePath = sourceDeclaration.GetVariablePath()
    sourceVariable = sourceContainer.AccessNamedElement(sourceVariablePath)
    targetVariablePath = sourceVariable.GetAbsolutePath()
    targetVariablePath.RemoveFirstIdentifier()
    targetVariablePath.MakeIdentifier()
    sourceConstant = sourceDeclaration.GetConstant()
    targetConstant = self._targetModel.NewConstant(sourceConstant)
    sourceDistribution = sourceDeclaration.GetDistribution()
    targetDistribution = self._FlattenStochasticExpression(sourceContainer, sourceDistribution)
    if targetDistribution==None:
      return S2MLFlattenerAction.FAILED
    targetDeclaration = self._targetModel.NewProbabilityDeclaration(targetVariablePath, targetConstant, targetDistribution, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenAssignment(self, action):
    sourceContainer = action.GetDatum(0)
    sourceAssignment = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    sourceVariablePath = sourceAssignment.GetVariablePath()
    sourceVariable = sourceContainer.AccessNamedElement(sourceVariablePath)
    targetVariablePath = sourceVariable.GetAbsolutePath()
    targetVariablePath.RemoveFirstIdentifier()
    targetVariablePath.MakeIdentifier()
    sourceExpression = sourceAssignment.GetExpression()
    targetExpression = self._FlattenStochasticExpression(sourceContainer, sourceExpression)
    if targetExpression==None:
      return S2MLFlattenerAction.FAILED
    targetAssignment = self._targetModel.NewAssignment(targetVariablePath, targetExpression, targetContainer)
    return S2MLFlattenerAction.COMPLETED

  def _FlattenObserver(self, action):
    sourceContainer = action.GetDatum(0)
    sourceObserver = action.GetDatum(1)
    targetContainer = action.GetDatum(2)
    sourceObserverPath = sourceObserver.GetPath()
    sourceObserver = sourceContainer.AccessNamedElement(sourceObserverPath)
    targetObserverPath = sourceObserver.GetAbsolutePath()
    targetObserverPath.RemoveFirstIdentifier()
    targetObserverIdentifier = targetObserverPath.ToString()    
    sourceDefinition = sourceObserver.GetDefinition()
    targetDefinition = self._FlattenStochasticExpression(sourceContainer, sourceDefinition)
    targetObserver = self._targetModel.NewObserver(targetObserverPath, targetDefinition, targetContainer)
    targetObserver.Rename(targetObserverIdentifier)
    return S2MLFlattenerAction.COMPLETED
