# Emmy version 1.0.0
# EmmyRewriter.py
# by Antoine Rauzy
# Copyrights (c) 2019 AltaRica Association
#
# This file is part of Emmy, a pedagogical stochastic simulator.
# Emmy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# Emmy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with Emmy.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Creator
# 3) Post-processing
# 4) Definition of references
# 3) Reordering of declarations

# 1) Imported modules
# -------------------

from S2MLCore import *

from EmmyModels import *

# 2) Creator
# ----------

class EmmyRewriter(S2MLRewriter):
  UNDEFINED = -1

  def __init__(self, engine):
    S2MLRewriter.__init__(self, engine)
    self._parameterIndex = 0
    self._variableIndex = 0
    self._probabilityDeclarationIndex = 0
    self._assignmentIndex = 0
    self._observerIndex = 0
    self._currentBlock = None
    self._errors = 0

# 3) Post-processing
# ------------------

  def PostProcessModel(self, model):
    self._errors = 0
    for element in model.GetElements():
      if element.GetType()==EmmyType.BLOCK:
        self.PostProcessBlock(element)
        if self._errors!=0:
          return self._errors
    return self._errors

  def PostProcessBlock(self, block):
    self.DefineReferencesInBlock(block)
    if self._errors!=0:
      return
    self.ReorderElementsInBlock(block)

# 4) Definition of references
# ---------------------------

  def DefineReferencesInBlock(self, block):
    previousBlock = self._currentBlock
    self._currentBlock = block
    for element in block.GetElements():
      if element.GetType()==EmmyType.BLOCK:
        self.DefineReferencesInBlock(element)
      elif element.GetType()==EmmyType.PARAMETER:
        self.DefineReferencesInParameterDeclaration(element)
      elif element.GetType()==EmmyType.VARIABLE:
        self.DefineReferencesInVariableDeclaration(element)
      elif element.GetType()==EmmyType.PROBABILITY_DECLARATION:
        self.DefineReferencesInProbabilityDeclaration(element)
      elif element.GetType()==EmmyType.ASSIGNMENT:
        self.DefineReferencesInAssignment(element)
      elif element.GetType()==EmmyType.OBSERVER:
        self.DefineReferencesInObserverDeclaration(element)
    self._currentBlock = previousBlock

  def DefineReferencesInParameterDeclaration(self, parameter):
    definition = parameter.GetDefinition()
    self.DefineReferencesInStochasticExpression(definition)

  def DefineReferencesInVariableDeclaration(self, variable):
    for name in variable.GetAttributeNames():
      expression = variable.GetAttribute(name)
      self.DefineReferencesInStochasticExpression(expression)

  def DefineReferencesInProbabilityDeclaration(self, declaration):
    distribution = declaration.GetDistribution()
    self.DefineReferencesInStochasticExpression(distribution)

  def DefineReferencesInAssignment(self, assignment):
    expression = assignment.GetExpression()
    self.DefineReferencesInStochasticExpression(expression)

  def DefineReferencesInObserverDeclaration(self, observer):
    definition = observer.GetDefinition()
    self.DefineReferencesInStochasticExpression(definition)

  def DefineReferencesInStochasticExpression(self, expression):
    if expression.GetType()==EmmyType.REFERENCE_EXPRESSION:
      path = expression.GetReferencePath()
      path.MakeIdentifier()
      element = self._currentBlock.AccessNamedElement(path)
      if element==None:
        self.RaiseError("In block %s, undefined reference %s" % (self._currentBlock.GetIdentifier(), path.ToString()))
        self._errors += 1
        return
      expression.SetReferredElement(element)
      return
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      return
    for child in expression.GetChildren():
      self.DefineReferencesInStochasticExpression(child)

# 3) Reordering of declarations
# -----------------------------

  def ReorderElementsInBlock(self, block):
    self._parameterIndex = 0
    self._variableIndex = 0
    self._probabilityDeclarationIndex = 0
    self._assignmentIndex = 0
    self._observerIndex = 0
    self.DefineIndicesInBlock(block)
    if self._errors!=0:
      return
    block.SortElements(EmmyRewriter._GetKeyOfElement)

  @staticmethod
  def _GetKeyOfElement(element):
    if element.GetType()==EmmyType.VARIABLE:
      return element.GetIndex()
    if element.GetType()==EmmyType.PROBABILITY_DECLARATION:
      return element.GetIndex()+10000
    if element.GetType()==EmmyType.PARAMETER:
      return element.GetIndex()+20000
    if element.GetType()==EmmyType.ASSIGNMENT:
      return element.GetIndex()+30000
    if element.GetType()==EmmyType.OBSERVER:
      return element.GetIndex()+40000
    return 0

  def DefineIndicesInBlock(self, block):
    for element in block.GetElements():
      if element.GetType()==EmmyType.VARIABLE:
        element.SetIndex(EmmyRewriter.UNDEFINED)
      elif element.GetType()==EmmyType.PROBABILITY_DECLARATION:
        element.SetIndex(EmmyRewriter.UNDEFINED)
      elif element.GetType()==EmmyType.PARAMETER:
        element.SetIndex(EmmyRewriter.UNDEFINED)
      elif element.GetType()==EmmyType.ASSIGNMENT:
        element.SetIndex(EmmyRewriter.UNDEFINED)
      elif element.GetType()==EmmyType.OBSERVER:
        element.SetIndex(EmmyRewriter.UNDEFINED)
    done = False
    while not done:
      done = True
      for element in block.GetElements():
        if element.GetType()==EmmyType.VARIABLE:
          unchanged = self.DefineIndicesInVariable(element)
          done = done and unchanged
        elif element.GetType()==EmmyType.PARAMETER:
          unchanged = self.DefineIndicesInParameter(element)
          done = done and unchanged
        elif element.GetType()==EmmyType.PROBABILITY_DECLARATION:
          unchanged = self.DefineIndicesInProbabilityDeclaration(element)
          done = done and unchanged
        elif element.GetType()==EmmyType.ASSIGNMENT:
          unchanged = self.DefineIndicesInAssignment(element)
          done = done and unchanged
        elif element.GetType()==EmmyType.OBSERVER:
          unchanged = self.DefineIndicesInObserver(element)
          done = done and unchanged
    for element in block.GetElements():
      if element.GetType()==EmmyType.PARAMETER and element.GetIndex()==EmmyRewriter.UNDEFINED:
        self.RaiseError("In block %s, absent or circular definition of parameter %s" % (block.GetIdentifier(), element.GetIdentifier()))
        self._errors += 1
      elif element.GetType()==EmmyType.VARIABLE and element.GetIndex()==EmmyRewriter.UNDEFINED:
        self.RaiseError("In block %s, absent or circular definition of variable %s" % (block.GetIdentifier(), element.GetIdentifier()))
        self._errors += 1
    return self._errors

  def DefineIndicesInVariable(self, variable):
    if variable.GetIndex()!=EmmyRewriter.UNDEFINED:
      return True
    if variable.GetCategory()==EmmyVariable.STATE:
      self._variableIndex += 1
      variable.SetIndex(self._variableIndex)
      return False
    return True

  def DefineIndicesInParameter(self, parameter):
    if parameter.GetIndex()!=EmmyRewriter.UNDEFINED:
      return True
    definition = parameter.GetDefinition()
    if self.AreAllIndicesOfReferencesInStochasticExpressionDefined(definition, True):
      self._parameterIndex += 1
      parameter.SetIndex(self._parameterIndex)
      return False
    return True

  def DefineIndicesInProbabilityDeclaration(self, declaration):
    if declaration.GetIndex()!=EmmyRewriter.UNDEFINED:
      return True
    distribution = declaration.GetDistribution()
    if self.AreAllIndicesOfReferencesInStochasticExpressionDefined(distribution, True):
      self._probabilityDeclarationIndex += 1
      declaration.SetIndex(self._probabilityDeclarationIndex)
      return False
    return True

  def DefineIndicesInAssignment(self, assignment):
    if assignment.GetIndex()!=EmmyRewriter.UNDEFINED:
      return True
    expression = assignment.GetExpression()
    if self.AreAllIndicesOfReferencesInStochasticExpressionDefined(expression, False):
      variable = assignment.GetVariable()
      self._variableIndex += 1
      variable.SetIndex(self._variableIndex)
      self._assignmentIndex += 1
      assignment.SetIndex(self._assignmentIndex)
      return False
    return True

  def DefineIndicesInObserver(self, observer):
    if observer.GetIndex()!=EmmyRewriter.UNDEFINED:
      return True
    definition = observer.GetDefinition()
    if self.AreAllIndicesOfReferencesInStochasticExpressionDefined(definition, False):
      self._observerIndex += 1
      observer.SetIndex(self._observerIndex)
      return False
    return True

  def AreAllIndicesOfReferencesInStochasticExpressionDefined(self, expression, includeParameters):
    if expression.GetType()==EmmyType.REFERENCE_EXPRESSION:
      element = expression.GetReferredElement()
      if element.GetType()==EmmyType.VARIABLE:
        return element.GetIndex()!=EmmyRewriter.UNDEFINED
      if element.GetType()==EmmyType.PARAMETER and includeParameters:
        return element.GetIndex()!=EmmyRewriter.UNDEFINED
      return True
    if expression.GetType()==S2MLType.CONSTANT_EXPRESSION:
      return True
    for child in expression.GetChildren():
      allDefined = self.AreAllIndicesOfReferencesInStochasticExpressionDefined(child, includeParameters)
      if not allDefined:
        return False
    return True
