# S2MLCore version 1.0.0
# S2MLDomain.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Domains

# 1) Imported modules
# -------------------

from S2MLConstant import *
from S2MLUniverse import *

# 2) Domains
# ----------

class S2MLDomain:
  BOOLEAN = 1
  INTEGER = 2
  REAL = 3
  SYMBOL = 4
  ENUMERATION = 5
  RANGE = 6
  REFERENCE = 7

  def __init__(self, type, universe):
    self._type = type
    self._universe = universe
    self._constants = []
    self._identifier = None

  @staticmethod
  def NewBoolean(universe):
    domain = S2MLDomain(S2MLDomain.BOOLEAN, universe)
    domain._constants.append(domain._universe.NewBoolean(False))
    domain._constants.append(domain._universe.NewBoolean(True))
    return domain
  @staticmethod
  def NewInteger(universe):
    domain = S2MLDomain(S2MLDomain.INTEGER, universe)
    return domain
  @staticmethod
  def NewReal(universe):
    domain = S2MLDomain(S2MLDomain.REAL, universe)
    return domain
  @staticmethod
  def NewSymbol(universe):
    domain = S2MLDomain(S2MLDomain.SYMBOL,universe)
    return domain
  @staticmethod
  def NewEnumeration(universe):
    domain = S2MLDomain(S2MLDomain.ENUMERATION, universe)
    return domain
  @staticmethod
  def NewRange(universe, lowerBound, upperBound):
    domain = S2MLDomain(S2MLDomain.RANGE, universe)
    for value in range(lowerBound, upperBound+1):
      constant = domain._universe.NewInteger(value)
      domain._constants.append(constant)
    return domain
  @staticmethod
  def NewReference(universe, identifier):
    domain = S2MLDomain(S2MLDomain.REFERENCE, universe)
    domain._identifier = identifier
    return domain

  def GetType(self):
    return self._type
  def SetType(self, type):
    return self._type
  def IsBoolean(self):
    return self._type==S2MLDomain.BOOLEAN
  def IsRange(self):
    return self._type==S2MLDomain.RANGE
  def IsInteger(self):
    return self._type==S2MLDomain.INTEGER
  def IsIntegral(self):
    return self._type==S2MLDomain.INTEGER or self._type==S2MLDomain.RANGE
  def IsReal(self):
    return self._type==S2MLDomain.REAL
  def IsNumerical(self):
    return self._type==S2MLDomain.INTEGER or self._type==S2MLDomain.RANGE or self._type==S2MLDomain.REAL
  def IsSymbol(self):
    return self._type==S2MLDomain.SYMBOL
  def IsEnumeration(self):
    return self._type==S2MLDomain.ENUMERATION
  def IsSymbolic(self):
    return self._type==S2MLDomain.SYMBOL or self._type==S2MLDomain.ENUMERATION
  def IsFinite(self):
    return self._type==S2MLDomain.BOOLEAN or self._type==S2MLDomain.ENUMERATION or self._type==S2MLDomain.SYMBOL
  def IsInfinite(self):
    return self._type==S2MLDomain.INTEGER or self._type==S2MLDomain.REAL or self._type==S2MLDomain.SYMBOL

  def GetMinimumValue(self):
    if self._type!=S2MLDomain.RANGE:
      return None
    if len(self._constants)==0:
      return None
    return self._constants[0]
  def GetMaximumValue(self):
    if self._type!=S2MLDomain.RANGE:
      return None
    if len(self._constants)==0:
      return None
    return self._constants[-1]
  def GetLowerBound(self):
    minimumValue = self.GetMinimumValue()
    if minimumValue==None:
      return None
    return minimumValue.GetValue()
  def GetUpperBound(self):
    maximumValue = self.GetMaximumValue()
    if maximumValue==None:
      return None
    return maximumValue.GetValue()

  def GetSize(self):
    if self.IsInfinite():
      return S2MLReal.PosInf()
    return len(self._constants)
  def Contains(self, constant):
    return constant in self._constants
  def GetIndex(self, constant):
    if not constant in self._constants:
      return -1
    return self._constants.index(constant)
  def LookForIndex(self, constant):
    index = 0
    for member in self._constants:
      if S2MLConstant.IsEq(member, constant):
        return index
      index += 1
    return -1
  def GetConstants(self):
    return self._constants
  def GetConstant(self, i):
    if i<0 or i>=len(self._constants):
      return None
    return self._constants[i]
  def AppendConstant(self, constant):
    self._constants.append(constant)
  def AppendConstants(self, constants):
    self._constants.extend(constants)
  def InsertConstant(self, index, constant):
    self._constants.insert(index, constant)
  def RemoveConstant(self, constant):
    self._constants.remove(constant)
  def PopConstant(self, index):
    self._constants.pop(index)

  def GetIdentifier(self):
    return self._identifier

  def Clone(self):
    clone = S2MLDomain(self._type, self._universe)
    clone._constants.extend(self._constants)
    clone._identifier = self._identifier
    return clone

