# S2MLCore version 1.0.0
# S2MLDistribution.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Points
# 3) Distributions

# 1) Imported modules
# -------------------

import sys
import math

# from S2MLDomains import *

# 2) Points
# ---------

class S2MLPoint:
  def __init__(self, x, y):
    self._x = x
    self._y = y

  def GetX(self):
    return self._x
  def GetY(self):
    return self._y

# 3) Distributions
# ----------------

class S2MLDistribution:
  def __init__(self):
    self._points = []

  def Reset(self):
    self._points = []

  def Copy(self):
    copy = S2MLDistribution()
    for point in self._points:
      pointCopy = S2MLPoint(point.GetX(), point.GetY())
      copy._points.append(pointCopy)
    return copy

  def GetNumberOfPoints(self):
    return len(self._points)
  def IsEmpty(self):
    return len(self._points)==0

  def IsCumulative(self):
    if self.GetNumberOfPoints()<=1:
      return True
    previousPoint = self._points[0]
    for index in range(1, self.GetNumberOfPoints()):
      currentPoint = self._points[index]
      if previousPoint.GetY()>currentPoint.GetY():
        return False
      prevousPoint = currentPoint
    return True

  def GetPoint(self, index):
    if index<0 or index>=self.GetNumberOfPoints():
      return None
    return self._points[index]
  def GetPoints(self):
    return self._points
  def GetLowestX(self, defaultValue = 0):
    if self.IsEmpty():
      return defaultValue
    point = self._points[0]
    return point.GetX()
  def GetHighestX(self, defaultValue = 0):
    if self.IsEmpty():
      return defaultValue
    point = self._points[-1]
    return point.GetX()
  def GetLowestY(self, defaultValue = 0):
    # REQUIRE: self.IsCumulative() 
    if self.IsEmpty():
      return defaultValue
    point = self._points[0]
    return point.GetY()
  def GetHighestY(self, defaultValue = 0):
    # REQUIRE: self.IsCumulative() 
    if self.IsEmpty():
      return defaultValue
    point = self._points[-1]
    return point.GetY()

  def LookForIndexOfPointAfterX(self, x):
    high = self.GetNumberOfPoints()
    if high==0:
      return 0
    point = self._points[0]
    if x<point.GetX():
      return 0
    point = self._points[-1]
    if x>point.GetX():
      return high
    low = 0
    high -= 1
    while low!=high:
      center = (low+high)//2
      point = self._points[center]
      if point.GetX()<=x:
        low = center+1
      else:
        high = center
    return low

  def LookForIndexOfPointAfterY(self, y):
    # REQUIRE: self.IsCumulative() 
    high = self.GetNumberOfPoints()
    if high==0:
      return 0
    point = self._points[0]
    if y<point.GetY():
      return 0
    point = self._points[-1]
    if y>point.GetY():
      return high
    low = 0
    high -= 1
    while low!=high:
      center = (low+high)//2
      point = self._points[center]
      if point.GetY()<=y:
        low = center+1
      else:
        high = center
    return low

  def AddPoint(self, x, y):
    index = self.LookForIndexOfPointAfterX(x)
    point = S2MLPoint(x, y)
    self._points.insert(index, point)
    return point
  def AddPairs(self, pairs):
    pairs.sort(key=lambda x: x[0])
    for pair in pairs:
      point = S2MLPoint(pair[0], pair[1])
      self._points.append(point)

  def GetXStepwise(self, y):
    # REQUIRE: self.IsCumulative() 
    if self.IsEmpty():
      return None
    index = self.LookForIndexOfPointAfterY(y)
    if index!=0:
      index -= 1
    point = self._points[index]
    return point.GetX()
   
  def GetXInterpolated(self, y):
    # REQUIRE: self.IsCumulative() 
    if self.IsEmpty():
      return None
    high = self.LookForIndexOfPointAfterY(y)
    if high==0:
      point = self._points[0]
      return point.GetX()
    low = high - 1
    if high==self.GetNumberOfPoints():
      point = self._points[low]
      return point.GetX()
    lowPoint = self._points[low]
    highPoint = self._points[high]
    x = lowPoint.GetX() + (highPoint.GetX()-lowPoint.GetX()) * (y - lowPoint.GetY())/(highPoint.GetY() - lowPoint.GetY())
    return x

  def GetYStepwise(self, x):
    if self.IsEmpty():
      return None
    index = self.LookForIndexOfPointAfterX(x)
    if index!=0:
      index -= 1
    point = self._points[index]
    return point.GetY()
   
  def GetYInterpolated(self, x):
    if self.IsEmpty():
      return None
    high = self.LookForIndexOfPointAfterX(x)
    if high==0:
      point = self._points[0]
      return point.GetY()
    low = high - 1
    if high==self.GetNumberOfPoints():
      point = self._points[low]
      return point.GetY()
    lowPoint = self._points[low]
    highPoint = self._points[high]
    y = lowPoint.GetY() + (highPoint.GetY()-lowPoint.GetY()) * (x - lowPoint.GetX())/(highPoint.GetX() - lowPoint.GetX())
    return y

  def PickUpXStepwise(self, z):
    # REQUIRE: self.IsCumulative() 
    # REQUIRE: 0 <= z <= 1
    if self.IsEmpty():
      return None
    lowestY = self.GetLowestY()
    highestY = self.GetHighestY()
    y = (z - lowestY)/(highestY - lowestY)
    return self.GetXStepwise(y)

  def PickUpXInterpolated(self, z):
    # REQUIRE: self.IsCumulative() 
    # REQUIRE: 0 <= z <= 1
    if self.IsEmpty():
      return None
    lowestY = self.GetLowestY()
    highestY = self.GetHighestY()
    y = (z - lowestY)/(highestY - lowestY)
    return self.GetXInterpolated(y)

  def PickUpYStepwise(self, z):
    # REQUIRE: 0 <= z <= 1
    if self.IsEmpty():
      return None
    lowestX = self.GetLowestX()
    highestX = self.GetHighestX()
    x = (z - lowestX)/(highestX - lowestX)
    return self.GetYStepwise(x)

  def PickUpYInterpolated(self, z):
    # REQUIRE: 0 <= z <= 1
    if self.IsEmpty():
      return None
    lowestX = self.GetLowestX()
    highestX = self.GetHighestX()
    x = (z - lowestX)/(highestX - lowestX)
    return self.GetYInterpolated(x)