# S2MLCore version 1.0.0
# S2MLIndicator.py
# by Antoine Rauzy
# Copyrights (c) 2018 AltaRica Association
#
# This file is part of S2MLCore.
# S2MLCore is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# S2MLCore is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with S2MLCore.  If not, see <http://www.gnu.org/licenses/>.

# Table of Contents
# -----------------
# 1) Imported modules
# 2) Indicators

# 1) Imported modules
# -------------------

import random

from S2MLDomains import *

from S2MLBin import *
from S2MLIndicatorProfile import *

# 2) Indicators
# -------------

class S2MLIndicator:
  MAXIMUM_NUMBER_OF_QUANTILES = 100
  DEFAULT_NUMBER_OF_QUANTILES = 10

  MAXIMUM_NUMBER_OF_POINTS = 100
  DEFAULT_NUMBER_OF_POINTS = 10

  MAXIMUM_NUMBER_OF_BINS = 1000
  DEFAULT_NUMBER_OF_BINS = 1000
  MAXIMUM_NUMBER_OF_STORED_VALUES = 10000
  DEFAULT_NUMBER_OF_STORED_VALUES = 10000
  MINIMUM_RATIO_NUMBER_OF_STORED_VALUES_NUMBER_OF_BINS = 10

  MAXIMUM_NUMBER_OF_SAMPLES = 1000
  DEFAULT_NUMBER_OF_SAMPLES = 20
  DEFAULT_SAMPLING_RATIO = 0.5

  def __init__(self, profile, domain):
    self._profile = profile
    self._domain = domain

    # Main measures
    if self._domain.IsNumerical():
      self._numberOfMeasures = 1
    elif self._domain.IsBoolean():
      self._numberOfMeasures = 1
    elif self._domain.IsEnumeration():
      self._numberOfMeasures = self._domain.GetSize()
    self._measures = [S2MLBin() for m in range(0, self._numberOfMeasures)]

    # Bins & Stored Values
    # The current version fixes the number of bins and stored values, for the sake of simplicity
    if self._domain.IsNumerical() and (self._profile.GetQuantiles() or self._profile.GetDistribution() or  self._profile.GetCumulativeDistribution() or self._profile.GetBins()):
      self._numberOfBins = S2MLIndicator.DEFAULT_NUMBER_OF_BINS
      self._bins = [S2MLBin() for b in range(0, self._numberOfBins)]
      self._numberOfStoredValues = S2MLIndicator.DEFAULT_NUMBER_OF_STORED_VALUES
      self._storedValues = [0 for v in range(0, self._numberOfStoredValues)]
      self._storedValueIndex = 0
    else:
      self._numberOfStoredValues = 0
      self._storedValues = []
      self._storedValueIndex = 0
      self._numberOfBins = 0
      self._bins = []

    # Samples
    if self._profile.GetBootstrapping():
      self._numberOfSamples = self._profile.GetNumberOfSamples()
      if self._numberOfSamples>S2MLIndicator.MAXIMUM_NUMBER_OF_SAMPLES:
        self._numberOfSamples = S2MLIndicator.MAXIMUM_NUMBER_OF_SAMPLES
      self._samplingRatio = self._profile.GetSamplingRatio()
      if self._samplingRatio<=0 or self._samplingRatio>=1.0:
        self._samplingRatio = S2MLIndicator.DEFAULT_SAMPLING_RATIO
      self._samples = [[S2MLBin() for m in range(0, self._numberOfMeasures)] for s in range(0, self._numberOfSamples)]
      self._bootstrappedMeasures = [S2MLBin() for m in range(0, self._numberOfMeasures)]
      self._CreateSampleMask2()
    else:
      self._numberOfSamples = 0
      self._samplingRatio = 0.0
      self._samples = []
      self._bootstrappedMeasures = []
      self._sampleMask = []

  def _CreateSampleMask1(self):
    self._sampleMask = [[] for sampleIndex in range(0, self._numberOfSamples)]
    for valueIndex in range(0, self._numberOfSamples):
      sampleList = self._sampleMask[valueIndex]
      for sampleIndex in range(0, self._numberOfSamples):
        if random.uniform(0.0, 1.0)<=self._samplingRatio:
          sampleList.append(sampleIndex)

  def _CreateSampleMask2(self):
    self._sampleMask = [None for sampleIndex in range(0, self._numberOfSamples)]
    self._numberOfSampledValues = int(math.floor(self._numberOfSamples*self._samplingRatio))
    for valueIndex in range(0, self._numberOfSamples):
      sampleList = [index for index in range(0, self._numberOfSamples)]
      while len(sampleList)>self._numberOfSampledValues:
        index = random.choice(sampleList)
        sampleList.remove(index)
      self._sampleMask[valueIndex] = sampleList

  def GetProfile(self):
    return self._profile
  def GetDomain(self):
    return self._domain

  def Reset(self):
    for m in range(0, self._numberOfMeasures):
      self._measures[m].Reset()
    for b in range(0, self._numberOfBins):
      self._bins[b].Reset()
    for s in range(0, self._numberOfSamples):
      for m in range(0, self._numberOfMeasures):
        self._samples[s][m].Reset()
    self._storedValueIndex = 0

  def GetMeasure(self, index=0):
    if index<0 or index>=self._numberOfMeasures:
      return None
    return self._measures[index]

  def GetNumberOfBins(self):
    return self._numberOfBins
  def GetNumberOfStoredValues(self):
    return self._numberOfStoredValues

  def GetBin(self, index):
    if index<0 or index>=self._numberOfBins:
      return None
    return self._bins[index]

  def GetNumberOfSamples(self):
    return self._numberOfSamples
  def GetSamplingRatio(self):
    return self._samplingRatio
  def GetNumberOfSampledValues(self):
    return self._numberOfSampledValues

  def GetSample(self, sampleIndex, valueIndex):
    if sampleIndex<0 or sampleIndex>=self._numberOfSamples or valueIndex<0 or valueIndex>=self._numberOfMeasures:
      return None
    return self._samples[sampleIndex][valueIndex]

  def AddValue(self, constant):
    # Main measure(s)
    if self._domain.IsNumerical():
      value = constant.GetValue()
      measure = self._measures[0]
      measure.AddValue(value)
    elif self._domain.IsBoolean():
      value = 1.0 if constant.IsTrue() else 0.0
      measure = self._measures[0]
      measure.AddValue(value)
    elif self._domain.IsEnumeration():
      value = self._domain.LookForIndex(constant)
      for m in range(0, self._numberOfMeasures):
        measure = self._measures[m]
        measure.AddValue(1 if m==value else 0)
    # Bins
    if self._numberOfStoredValues>0:
      self._storedValues[self._storedValueIndex] = value
      self._storedValueIndex += 1
      if self._storedValueIndex==self._numberOfStoredValues:
        self.FlushStoredValues()
    # Bootstrapping
    if self._sampleMask!=[]:
      valueIndex = self._measures[0].GetNumberOfValues()%self._numberOfSamples
      samplingList = self._sampleMask[valueIndex]
      for sampleIndex in samplingList:
        sample = self._samples[sampleIndex]
        if self._domain.IsNumerical():
          measure = sample[0]
          measure.AddValue(value)
        elif self._domain.IsBoolean():
          measure = sample[0]
          measure.AddValue(value)
        elif self._domain.IsEnumeration():
          for m in range(0, self._numberOfMeasures):
            measure = sample[m]
            measure.AddValue(1 if m==value else 0)

  def FlushStoredValues(self):
    if self._storedValueIndex==0:
      return
    if self._numberOfBins>0:
      if self._storedValueIndex==self._numberOfStoredValues:
        sortedValues = self._storedValues
      else:
        sortedValues = self._storedValues[0: self._storedValueIndex]
      sortedValues.sort()
      ratio = float(self._storedValueIndex)/float(self._numberOfBins)
      valueIndex = 0
      binIndex = 0
      bin = self.GetBin(binIndex)
      border = ratio
      for valueIndex in range(0, self._storedValueIndex):
        if valueIndex>border:
          binIndex += 1
          bin = self.GetBin(binIndex)
          border += ratio
        bin.AddValue(sortedValues[valueIndex])
    self._storedValueIndex = 0

  def GetBootstrappedMeasure(self, index=0):
    if self._bootstrappedMeasures==[] or index<0 or index>=self._numberOfMeasures:
      return None
    return self._bootstrappedMeasures[index]
  def CalculateBootstrappedMeasures(self):
    if self._bootstrappedMeasures==[]:
      return
    self.FlushStoredValues()
    for m in range(0, self._numberOfMeasures):
      self._bootstrappedMeasures[m].Reset()
    for s in range(0, self._numberOfSamples):
      for m in range(0, self._numberOfMeasures):
        self._bootstrappedMeasures[m].AddValue(self._samples[s][m].GetMean())

  def ExtractQuantiles(self):
    numberOfValues = self._measures[0].GetNumberOfValues()
    if numberOfValues==0 or self._storedValues==[]:
      return []
    numberOfQuantiles = self._profile.GetNumberOfQuantiles()
    if numberOfQuantiles<=0:
      numberOfQuantiles = S2MLIndicator.DEFAULT_NUMBER_OF_QUANTILES
    if numberOfQuantiles>S2MLIndicator.MAXIMUM_NUMBER_OF_QUANTILES:
      numberOfQuantiles = S2MLIndicator.MAXIMUM_NUMBER_OF_QUANTILES
    if numberOfValues<=numberOfQuantiles:
      return []
    # -------
    if numberOfValues<self._numberOfStoredValues:
      sortedValues = self._storedValues[0: self._storedValueIndex]
      sortedValues.sort()
      quantiles = []
      ratio = float(numberOfValues)/float(numberOfQuantiles)
      for quantileIndex in range(1, numberOfQuantiles+1):
        valueIndex = int(math.floor(quantileIndex*ratio))-1
        quantile = (float(quantileIndex)/float(numberOfQuantiles), sortedValues[valueIndex])
        quantiles.append(quantile)
      return quantiles
    # -------
    self.FlushStoredValues()
    if numberOfQuantiles>self._numberOfBins:
      numberOfQuantiles = self._numberOfBins
    quantiles = []
    ratio = float(numberOfValues)/float(numberOfQuantiles)
    binIndex = 0
    bin = self.GetBin(binIndex)
    cumulatedNumberOfValues = 0
    for quantileIndex in range(1, numberOfQuantiles+1):
      expectedNumberOfValues = int(math.floor(quantileIndex*ratio))
      while cumulatedNumberOfValues<expectedNumberOfValues:
        cumulatedNumberOfValues += bin.GetNumberOfValues()
        if binIndex==self._numberOfBins-1:
          break
        binIndex += 1
        bin = self.GetBin(binIndex)
      quantile = (float(quantileIndex)/float(numberOfQuantiles), bin.GetMean())
      quantiles.append(quantile)
    return quantiles

  def ExtractDistribution(self):
    numberOfValues = self._measures[0].GetNumberOfValues()
    if numberOfValues==0 or self._storedValues==[]:
      return []
    numberOfPoints = self._profile.GetNumberOfPoints()
    if numberOfPoints<=0:
      numberOfPoints = S2MLIndicator.DEFAULT_NUMBER_OF_POINTS
    if numberOfPoints>S2MLIndicator.MAXIMUM_NUMBER_OF_POINTS:
      numberOfPoints = S2MLIndicator.MAXIMUM_NUMBER_OF_POINTS
    if numberOfValues<=numberOfPoints:
      return []
    # -------
    if numberOfValues<self._numberOfStoredValues:
      sortedValues = self._storedValues[0: self._storedValueIndex]
      sortedValues.sort()
      minimumValue = sortedValues[0]
      maximumValue = sortedValues[self._storedValueIndex-1]
      points = []
      pointIndex = 1
      ratio = float(abs(maximumValue-minimumValue))/float(numberOfPoints)
      valueIndex = 0
      border = minimumValue
      for pointIndex in range(1, numberOfPoints+1):
        border += ratio
        previousValueIndex = valueIndex
        while valueIndex<numberOfValues and sortedValues[valueIndex]<=border:
          valueIndex += 1
        intensity = float(valueIndex-previousValueIndex)/float(numberOfValues)
        point = (border, intensity)
        points.append(point)
      return points
    # -------
    self.FlushStoredValues()
    if numberOfPoints>self._numberOfBins:
      numberOfPoints = self._numberOfBins
    firstBin = self.GetBin(0)
    lastBin = self.GetBin(self._numberOfBins-1)
    minimumValue = firstBin.GetMinimumValue()
    maximumValue = lastBin.GetMaximumValue()
    ratio = float(abs(maximumValue-minimumValue))/float(numberOfPoints)
    points = []
    pointIndex = 1
    binIndex = 0
    bin = self.GetBin(binIndex)
    binValue = bin.GetMean()
    border = minimumValue
    for pointIndex in range(1, numberOfPoints+1):
      numberOfAccumulatedValues = 0
      border += ratio
      if binIndex<self._numberOfBins:
        while binValue<=border:
          numberOfAccumulatedValues += bin.GetNumberOfValues()
          binIndex += 1
          if binIndex==self._numberOfBins:
            break
          bin = self.GetBin(binIndex)
          binValue = bin.GetMean()
      intensity = float(numberOfAccumulatedValues)/float(numberOfValues)
      point = (border, intensity)
      points.append(point)
    return points

  def ExtractCumulativeDistribution(self):
    numberOfValues = self._measures[0].GetNumberOfValues()
    if numberOfValues==0 or self._storedValues==[]:
      return []
    numberOfPoints = self._profile.GetNumberOfPoints()
    if numberOfPoints<=0:
      numberOfPoints = S2MLIndicator.DEFAULT_NUMBER_OF_POINTS
    if numberOfPoints>S2MLIndicator.MAXIMUM_NUMBER_OF_POINTS:
      numberOfPoints = S2MLIndicator.MAXIMUM_NUMBER_OF_POINTS
    if numberOfValues<=numberOfPoints:
      return []
    # -------
    if numberOfValues<self._numberOfStoredValues:
      sortedValues = self._storedValues[0: self._storedValueIndex]
      sortedValues.sort()
      minimumValue = sortedValues[0]
      maximumValue = sortedValues[self._storedValueIndex-1]
      points = []
      pointIndex = 1
      ratio = float(abs(maximumValue-minimumValue))/float(numberOfPoints)
      valueIndex = 0
      border = minimumValue
      for pointIndex in range(1, numberOfPoints+1):
        border += ratio
        while valueIndex<numberOfValues and sortedValues[valueIndex]<=border:
          valueIndex += 1
        intensity = float(valueIndex)/float(numberOfValues)
        point = (border, intensity)
        points.append(point)
      return points
    # -------
    self.FlushStoredValues()
    if numberOfPoints>self._numberOfBins:
      numberOfPoints = self._numberOfBins
    firstBin = self.GetBin(0)
    lastBin = self.GetBin(self._numberOfBins-1)
    minimumValue = firstBin.GetMinimumValue()
    maximumValue = lastBin.GetMaximumValue()
    ratio = float(abs(maximumValue-minimumValue))/float(numberOfPoints)
    points = []
    pointIndex = 1
    binIndex = 0
    bin = self.GetBin(binIndex)
    binValue = bin.GetMean()
    numberOfAccumulatedValues = 0
    border = minimumValue
    for pointIndex in range(1, numberOfPoints+1):
      border += ratio
      if binIndex<self._numberOfBins:
        while binValue<=border:
          numberOfAccumulatedValues += bin.GetNumberOfValues()
          binIndex += 1
          if binIndex==self._numberOfBins:
            break
          bin = self.GetBin(binIndex)
          binValue = bin.GetMean()
      intensity = float(numberOfAccumulatedValues)/float(numberOfValues)
      point = (border, intensity)
      points.append(point)
    return points
